/*
 utility/interpreter.c

  Created: 2008.08.11
  Revised: 2010.04.04

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "../types.h"
#include "ftrig.h"
#include "interpreter.h"

#ifdef env68k
 #pragma segment Interpreter

 #define native_double			long double
#else
 #define native_double			double
#endif

native_double sqrt(native_double);
native_double pow(native_double, native_double);

/* ---- typedefs */

typedef unsigned char		char_type, op_type;

/* ---- constants */

enum
{/* op_type types */
	_op_none,
	_op_expr,	/* to signal coming after any expression */
 /* unary */
	_op_ldw,	/* @	[set (future): `=>`] */
	_op_sqr,	/* ~/	[or]	CHR_RA */

	_op_not,	/* ~ */
	_op_bno,	/* ! */
	_op_neg,	/* -	[or]	CHR_MN */
	_op_abs,	/* + */
 /* binary / dyadic */
	_op_and,	/* & */
	_op_ban,	/* && */
	_op_ior,	/* | */
	_op_bor,	/* || */

	_op_eor,	/* ^ */
	_op_asr,	/* >>> */
	_op_lsr,	/* >> */
	_op_lsl,	/* << */

	_op_add,	/* + */
	_op_sub,	/* - */
	_op_mul,	/* * */
	_op_pow,	/* ** */

	_op_dvu,	/* / */
	_op_dvs,	/* //	[or]	CHR_DV */
	_op_mdu,	/* % */
	_op_mds,	/* %% */

	_op_clt,	/* < */
	_op_cge,	/* >= */
	_op_cgt,	/* > */
	_op_cle,	/* <= */

	_op_ceq,	/* = */
	_op_cne,	/* !=	[or]	CHR_NE */
	_op_cid,	/* ==	(identical) */
	_op_cni		/* <>	(not identical) */
};

#define CHR_HX				'$'
#ifdef mac
 #define CHR_NE				'\255' /* inequal */
 #define CHR_LE				'\262' /* lesser or equal */
 #define CHR_GE				'\263' /* greater or equal */
 #define CHR_RA				'\303' /* radical (square root) */
 #define CHR_DS				'\312' /* double-space / non-breaking space */
 #define CHR_MN				'\320' /* minus "dash" */
 #define CHR_DV				'\326' /* divide */
#else if defined(iso8859_1)
 //no NE
 //no official LE
 //no official GE
 //no RA
 #define CHR_DS				'\240' /* double-space / non-breaking space */
 #define CHR_MN				'\226' /* minus "dash" */
 #define CHR_DV				'\367' /* divide */
#endif

/* ---- macros */

#define CASE_NUMERIC			case'0':case'1':case'2':case'3':case'4': \
	case'5':case'6':case'7':case'8':case'9':

#define CASE_ALPHABETIC			case'A':case'B':case'C':case'D':case'E': \
	case'F':case'G':case'H':case'I':case'J':case'K':case'L':case'M':case'N': \
	case'O':case'P':case'Q':case'R':case'S':case'T':case'U':case'V':case'W': \
	case'X':case'Y':case'Z':	case'a':case'b':case'c':case'd':case'e': \
	case'f':case'g':case'h':case'i':case'j':case'k':case'l':case'm':case'n': \
	case'o':case'p':case'q':case'r':case's':case't':case'u':case'v':case'w': \
	case'x':case'y':case'z':

/* case-only; hyphen and plus account for binary and unary functions */
/*#define CASE_UNARY_OPERATOR case'!':case'@':case'~':*/
#define CASE_BINARY_OPERATOR case'%':case'&':case'*':case'/': \
	case'<':case'=':case'>':case'^':case'|':

#ifdef CHR_DS
 #define skipwsp(s,e)			do{char_type c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_DS)!=0) \
		{--(s); break;}}while(s<e)
 #define rskipwsp(s,f)			do{char_type c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_DS)!=0) \
		{++(s); break;}}while(s>=f)
#else
 #define skipwsp(s,e)			do{char_type c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {--(s); break;}}while(s<e)
 #define rskipwsp(s,f)			do{char_type c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {++(s); break;}}while(s>=f)
#endif

#define skipint(s,e)			do{if((char_type)(*(s)++-'0')>'9'-'0') \
	{--s; break;}} while(s<e)

/* ---- globals */

#if defined(mac) || defined(isostd)
static const uint8	uppertable[] =
{
	0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
	0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,
	0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
	0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
	0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,
	0x60,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
	0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x7B,0x7C,0x7D,0x7E,0x7F,
#ifdef mac
	0x80,0x81,0x82,0x83,0x84,0x85,0x86,0xE7,0xCB,0xE5,0x80,0xCC,0x81,0x82,0x83,0xE9,
	0xE6,0xE8,0xEA,0xED,0xEB,0xEC,0x84,0xEE,0xF1,0xEF,0x85,0xCD,0xF2,0xF4,0xF3,0x86,
	0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
	0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xC6,0xB7,0xB8,0xB8,0xBA,0xBB,0xBC,0xBD,0xAE,0xAF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCE,
	0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD9,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
	0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,
	0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF
#elif defined(iso8859_1)
	0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x90,
	0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x8C,0x9D,0x9E,0x9F,
	0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
	0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
	0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
	0xF0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xFD,0xDE,0x9F,
#endif
};

/* ---- system code */

int strcasecmp(const char *a,
	const char *b)
{
	const uint8	*t = uppertable;
	uint8		c, d;

	if (a==NULL || b==NULL)
		return 1+(((unsigned int)-1)>>1);
	do
		if ((c=t[*((uint8*)a)++])==0)
			return -*b;
	while(c==(d=t[*((uint8*)b)++]));

	return c-d;
}

int strncasecmp(const char *a,
	const char *b,
	size_t z)
{
	const uint8	*t = uppertable;
	uint8		c, d;

	if (a==NULL || b==NULL)
		return 1+(((unsigned int)-1)>>1);
	if ((signed)z<=0) /* into unmapped space? */
		return 0;
	while((c=t[*((uint8*)a)++])==(d=t[*((uint8*)b)++]) && --z!=0) ;

	return c-d;
}
#endif

/* extended standard */

const char *readanyinteger(const char *s, long z, long int *v)
{
	const char		*e;
	unsigned long int	r = 0, m = 0;
	char_type		c;

	if (s==0 || z<=0)
		return 0;
	e = z+s;
	c = *s++;
	if ((c+=-'-')==0)
	{
		if (s>=e)
			return 0;
		m = -1, c = *s++;
	} else if ((c+='-'-'+')==0)
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if ((char_type)((c+='+')-'0')>(char_type)('9'-'0'))
	{
#ifdef CHR_HX
		if (c==CHR_HX)
			goto hexadecimal;
#endif
		return 0;
	}
	if (c!='0')
	{
		goto decimal;
		do
		{
			c = *s++;
		decimal:
			c += -'0';
			if (c>(char_type)('9'-'0'))
				if ((c+='0'-'A')<=(char_type)('Z'-'A')
						|| (c+='A'-'a')<=(char_type)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			r = r*10+c;
		} while(s<e);
		goto end;
	}

	if (s>=e)
		goto end;
	c = *s++;
	if (c!='x' && c!='X')
	{
		if (c=='o' || c=='O')
		{ /* octal (starting-zero-only not supported here) */
			do
			{
				c = *s++, c += -'0';
				if (c>(char_type)('7'-'0'))
					if (c>(char_type)('9'-'7')
							|| (c+='0'-'A')<=(char_type)('Z'-'A')
							|| (c+='A'-'a')<=(char_type)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r = (r<<3)+c;
			} while(s<e);
			goto end;
		}
		if (c=='b' || c=='B')
		{ /* binary */
			do
			{
				c = *s++, c += -'0';
				if (c>(char_type)('1'-'0'))
					if (c>(char_type)('9'-'1')
							|| (c+='0'-'A')<=(char_type)('Z'-'A')
							|| (c+='A'-'a')<=(char_type)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r += r, r += c;
			} while(s<e);
			goto end;
		}
		goto decimal;
	}
hexadecimal:
	if (s>=e) /* only "0x"? */
		return 0;
	do
	{
		c = *s++, c += -'0';
		if (c<=(char_type)('9'-'0'))
			r = r*16+c;
		else if ((c+='0'-'A')<=(char_type)('Z'-'A')
				|| (c+='A'-'a')<=(char_type)('z'-'a'))
			if (c>(char_type)('F'-'A'))
				{s = 0; break;}
			else
				c += 10, r = r*16+c;
		else if ((c+='a'-'_')==0)
			{s = 0; break;}
		else
			{--s; break;}
	} while(s<e);
end:
	r += m, r^= m;
	if (v!=0)
		*v = r;

	return s;
}

const char *parse_number(const char *s,
	const char *e,
	double *_v,
	unsigned char r)
{
	native_double	fb; //can be a long double since this comes from an integer
	double		v = 0;
	int		u = 0, a = 0, b = 10, m = 0, x = 0;
	unsigned char	c, n = 0;

	skipwsp(s,e);

#if CHR_MN!='-'
	if ((c=*s++)=='-' || c==(unsigned char)CHR_MN)
#else
	if ((c=*s++)=='-')
#endif
	{
		n = 1;
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c=='+')
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c!='.' && (unsigned char)(c-'0')>(unsigned char)('9'-'0'))
		return 0; /* first must be a digit */

	if (c=='0' && s<e)
	{
		if ((c=*s-'X')==0 || (c+='X'-'x')==0)
			b = 16, u = 3, ++s, r = 0; /* fractionless hexadecimal */
		else if ((c+='x'-'h')==0)
			b = 16, u = 3, ++s; /* hexadecimal with fraction */
		else if ((c+='h'-'O')==0 || (c+='O'-'o')==0)
			b = 8, ++s; /* octal (starting-zero-only not supported here) */
		else if ((c+='o'-'b')==0)
			b = 2, ++s; /* binary */
		else if ((c+='b'-'d')==0)
			b = 10, r = 0, ++s; /* fractionless decimal */
		else if ((c+='d'-'z')==0)
			b = 36, u = 3, ++s; /* base 36 (all alphanumeric) */
		if (s>=e)
			return 0; /* must follow alphanumeric digits */
		c = *s++;
	}
	fb = b;
	goto start;
	do
	{
		c = *s++;
	start:
		if ((c+=-'0')<=(unsigned char)('9'-'0')) /* 0-9 */
		{
			if (c>=(unsigned char)b)
				return 0; /* digit invalid for this base number */
			u|= 0x8000;
			if (a>0)
				x = 10*x+c;
			else
				v = b*v+c, m += a;
		} else if ((c+='0'-'A')<='Z'-'A' || (c+='A'-'a')<='z'-'a' && (unsigned char)u>2)
		{
			if ((c+=10)>=(unsigned char)b)
				return 0; /* alphanumeric digit not valid in this base number */
			v = b*v+c, m += a, u|= 0x8000;
			continue;
		} else if ((c+='A'-'E')==0) /* Ee */
		{ /* can't have more than one `E' */
			if (a==1)
				return 0;
			a = 1;
		} else if ((c=s[-1]-'.')==0) /* . */
		{ /* can't have more than one point or during integer parsing (includes notation) */
			if (a!=0 || !r)
				return 0;
			a = -1;
#if CHR_MN!='-'
		} else if ((c+='.'-'-')==0 || c==(unsigned char)(CHR_MN-'-'))
#else
		} else if ((c+='.'-'-')==0)
#endif
		{ /* only one '-' valid (here) for notation; cannot come after `E+' */
			if (a<=0 || (u&1)!=0)
				return 0;
			u = 1;
		} else if ((c+='-'-'+')==0) /* + */
		{ /* only one '+' valid (here) for notation; cannot come after `E-' */
			if (a<=0 || (unsigned char)u!=0)
				return 0;
			u = 2;
		} else if ((c+='+'-'_')==0)
			return 0; /* can't directly follow an underscore */
		else /* everything else ends */
			break;
	} while(s<e);

	if ((u&0x8000)==0)
		return 0; /* no digit chars after start or notation ("...E+" | "...E-") */
	if ((unsigned char)u==1)
		m -= x;
	else
		m += x;
	if (m!=0)
		v *= pow(b,m);

	if (_v)
		*_v = n==0 ? v:-v;

	return s;
}

int makebasestring(void *outstring,
	const char *basechars,
	long value,
	unsigned base,
	int minlength,
	char signchar)
{
	const char	*internalbasechars =
		"0123456789ABCDEFGHIJKLMNOPQRSTUVXYZ@#$`!%&()*+,./:;<=>?[\\]^{|}~";
	unsigned char	store[32]; /* enough space to do binary */
	unsigned char	*s = sizeof(store)+store;
	unsigned	v = value;
	int		x = 0;

	if (base==0)
		return 0;
	/* SC generated erroneous code if `value` wasnt used in the main loop, */
	/* optimizing it out entirely as `v`, and nixing this conditional as a result */
	if ((signed)v<0 && signchar!=0)
		v = -v;
	if (basechars==0)
		basechars = internalbasechars;
	if (minlength<=0)
		minlength = 1;
	else if (minlength>sizeof(store))
		minlength = sizeof(store);

	do
	{
		unsigned	i = v;

		*--s = basechars[i-=(v/=base)*base];
	} while(++x<minlength || v!=0);
	minlength = x;

	if (outstring)
	{
		unsigned char	*d = outstring;

		if (signchar!=0 && value<0)
			*d++ = signchar, ++minlength;

		/* copy to output -- no function calls, unaligned compatible */
		if (x&1)
			*d++ = *s++;
		if (x&2)
			*d++ = *s++, *d++ = *s++;
		if ((x&=-4)!=0)
			do
				*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
			while((x-=4)!=0);
	}

	return minlength;
}

int parse_textloop(const char *t,
	unsigned a,
	void (*b)(unsigned a, unsigned char v))
{
	const uint8	*s;
	int		m = 0, w = 3, z = 0;
	uint8		c, q;

	if ((s=(const uint8*)t)==0)
		goto end;
	if ((q=*s++)!=0 && q!='"' && q!='\'')
		return 0;
	while((c=*s++)!=0)
	{
		if (m!=0)
			if (m<0)
				goto form;
			else if ((uint8)(c-'0')>(uint8)9)
			{
				if (w>3)
				{
					if ((c+=-'A')<=(uint8)('F'-'A')
							|| (c+='A'-'a')<=(uint8)('f'-'a'))
						{m = (m<<w)|c+10; continue;}
					if (m==0x800000)
						return ~(s-(const uint8*)t);
					c += 'a';
				}
			form:
				b(a++,m&0xFF); m = 0, ++z;
			} else
				{m = (m<<w)|(c&7); continue;}
		if (c==q)
			break;
		else if (c=='\\')
		{
			if ((c=*s++)==0)
				return ~(s-(const uint8*)t);
			else if ((uint8)(c-'0')<=(uint8)7)
				{m = 0x3800000|(c&7), w = 3; continue;}
			else if ((uint8)(c-10)<=(uint8)3)
				continue;

			switch(c)
			{
			case 't': c = 9; break;
			case 'n': c = 10; break;
			case 'v': c = 11; break;
			case 'f': c = 12; break;
			case 'r': c = 13; break;

			case 'x':
				m = 0x800000, w = 4;
				continue;
			case 'p':
				c = strlen((const char*)s);
				if (s[--c]!=q)
					++c;
			}
		}
		b(a++,c); ++z;
	}
end:
	return z;
}

const char *parse_skipwsp(const char *s, const char *e)
{
	if (s==0)
		return 0;
	skipwsp(s,e);

	return s;
}

const char *parse_wordbreak(const char *s, const char *e)
{
	if (s==0)
		return 0;
	skipwsp(s,e);
	if (s==e)
		return s;
	do
	{
		char_type	c = *s++;

		if ((c+=-'_')!=0 && (c+='_'-'0')>(char_type)('9'-'0')
				&& (c+='0'-'A')>(char_type)('Z'-'A')
				&& (c+='A'-'a')>(char_type)('z'-'a'))
			{--s; break;}
	} while(s<e);

	return s;
}

const char *parse_rwordbreak(const char *s, const char *f)
{
	if (s==0)
		return 0;
	rskipwsp(s,f);
	if (s==f)
		return s;
	do
	{
		char_type	c = *--s;

		if ((c+=-'_')!=0 && (c+='_'-'0')>(char_type)('9'-'0')
				&& (c+='0'-'A')>(char_type)('Z'-'A')
				&& (c+='A'-'a')>(char_type)('z'-'a'))
			{++s; break;}
	} while(s>=f);

	return s;
}

static const char *quotedtext_scan(const char *s, const char *e, char_type v)
{
	if (e<=s)
		return 0;
	do
	{
		char_type	c = *s++;

		if (c=='\\') /* escape; skip next char */
			++s;
		else if (c==v)
			return s;
	} while(s<e);

	return 0; /* no closing quote */
}

/* ---- evaluation! */

/* would be true if evalvalue is floating-point */
#define EVAL_FLOATING				((((evalvalue)1)/2)*2!=0)

static evalvalue epow(evalvalue a,
	evalvalue b)
{
	if (EVAL_FLOATING)
		return pow(a,b);
	else if (b<=0)
		return a==0 ? EVAL_HUGEVAL:b==0 ? 1:++b ? 0:1/a;
	else if (a==1 || --b==0)
		return a;
	else if (b>31)
		return a==0 ? 0:++a ? EVAL_HUGEVAL:(~b&1)*2-1;
	else
	{
		int64		v;
		evalvalue	m = a;

		do
		{
			(void)muls64s32(v,a,m);
			if ((a=int64lo(v))>>32!=int64hi(v))
				return EVAL_HUGEVAL; /* overflow */
		} while(--b!=0);
	}

	return a;
}

static evalvalue esqrt(evalvalue a)
{
	if (EVAL_FLOATING)
		return sqrt(a);
	else
		return usqrt(a);
}

static evalvalue evaluate_unary(char_type c,
	evalvalue v,
	int n, /* if zero, treat NOT as sqrt */
	evalvalue (*evalindir)(evalvalue))
{
	if ((c+=-'@')==0) /* read address */
	{
		if (evalindir)
			v = evalindir(v);
	} else if ((c+='@'-'~')==0) /* binary not */
		v = n ? ~v:esqrt(v);
	else if ((c+='~'-'!')==0) /* boolean not */
		v = !v;
#ifdef CHR_MN
	else if (c==(char_type)(CHR_MN-'!')) /* minus-negate */
		v = -v;
#endif
#ifdef CHR_RA
	else if (c==(char_type)(CHR_RA-'!')) /* square root */
		v = esqrt(v);
#endif
	else if (c==(char_type)('+'-'!')) /* plus-absolute value */
	{
		if (v<0)
			v = -v;
	} else if (c==(char_type)('-'-'!')) /* hyphen-negate */
		v = -v;
	/* ignore anything else (should have already processed it) */

	return v;
}

static evalvalue evaluate_binary(op_type c,
	evalvalue a,
	evalvalue b)
{
	switch(c)
	{
	case _op_and: a&= b; break; /* binary AND */
	case _op_ban: a = a && b; break; /* boolean AND */
	case _op_ior: a|= b; break; /* binary OR */
	case _op_bor: a = a || b; break; /* boolean OR */
	case _op_eor: a^= b; break; /* binary Exclusive-OR */
	case _op_asr: a>>= b; break; /* arithmetic shift right */
	case _op_lsr: a = (unsigned)a>>(unsigned)b; break; /* logical shift right */
	case _op_lsl: a = (unsigned)a<<(unsigned)b; break; /* logical shift left */
	case _op_add: a+= b; break; /* add */
	case _op_sub: a-= b; break; /* subtract */
	case _op_mul: a = (unsigned)a*(unsigned)b; break; /* multiply */
	case _op_pow: a = epow(a,b); break; /* exponentiation */
	case _op_dvu: a = b==0 ? -1U:(unsigned)a/(unsigned)b; break; /* u-div */
	case _op_dvs: a = b==0 ? EVAL_HUGEVAL:a/b; break; /* signed division */
	case _op_mdu: a = b==0 ? 0:(unsigned)a%(unsigned)b; break; /* u-mod */
	case _op_mds: a = b==0 ? 0:a%b; break; /* signed modulus */
	case _op_clt: a = a<b; break; /* lesser */
	case _op_cge: a = a>=b; break; /* greater or equal */
	case _op_cgt: a = a>b; break; /* greater */
	case _op_cle: a = a<=b; break; /* lesser or equal */
	case _op_ceq: case _op_cid: a = a==b; break; /* equal/identical */
	case _op_cne: case _op_cni: a = a!=b; break; /* not equal/identical */
	default: a = b;
	}

	return a;
}

//the necessary update has been sitting (in terms of this srcfile) for a whole year:
// have the frame be a stack of codes delimited by dyadic ops
// it would eliminating pointers, be alignment for `v` would be tricky...
typedef struct evalframe/*(13+)*/
{ /* for whereever there is a parenthesis... */
	evalvalue	v; /* value */
	op_type		o; /* opcode (see enum) */
	const char	*u; /* first unary operator address (special case) */
	const char	*e; /* end-expression address (start of the next expression) */
} evalframe;

/* evaluate_chunk evaluates any unary operations, then the main binary operator */
/* it expects the current and next value in each frame to be valid */
static int evaluate_chunk(evalframe *a,
	evalvalue b,
	evalvalue (*evalindir)(evalvalue))
{
	const char	*s;
	int		n = 1; /* stupid workaround for ASCII sqrt... */

	/*if (a->v==EVAL_HUGEVAL)
		return -1;*/
	if (a->o==_op_none)
		{a->v = b; return 0;} /* no preceeding expression; copy */

	/* unary */
	if ((s=a->e)!=0)
		while(--s>=a->u)
		{
			if (*s=='/' && (unsigned)(a->o-_op_dvu)>1)
				n = 0;
			b = evaluate_unary(*s,b,n,evalindir);
			if (*s=='~')
				n = 1;
		}

	/* binary */
	a->v = evaluate_binary(a->o,a->v,b);

	return 1;
}

#define EVAL_STACKSIZE				32

static const char *evaluate_parameter(const char *str,
	const char *e,
	evalvalue *v,
	const char *(*getvalue)(const char *s, const char *e, long int *i, uint8 *u),
	uint8 *u,
	evalvalue (*evalindir)(evalvalue))
{
	evalframe	xstack[EVAL_STACKSIZE+1]; /* +1 for local operators */

	evalframe	*a = xstack;
	const char	*s = str, *p;
	long int	i;
	int		x = 0, n = 0;
	op_type		c, o;

	/* clear frame */
	a->v = 0;
	a->o = o = _op_none;
	a->u = (const char*)-1;
	a->e = 0;

	do
	{
		switch((char_type)*s) /* try not to get stuck; make sure we advance */
		{
		case ' ':
		whitespace: /*o |= 0x80;*/
			skipwsp(s,e);
			break;
		case '#':
			a->e = s;
			if (++s>=e) /* ignore first '#' for numeric */
				{--s; goto end;}
#ifdef CHR_HX
		case CHR_HX:
#endif
		CASE_NUMERIC
			if (a->e==0)
				a->e = s;
			/* non-integer or another alphanumeric */
			if ((p=readanyinteger(s,e-s,&i))==0
					|| o==_op_expr)
				{s = a->e; goto end;} 
			s = p, ++n;

			if (o==_op_none) /* no operator/expression; set */
				a->v = i;
			else /* evaluate */
				evaluate_chunk(a,i,evalindir);

			a->o = o = _op_expr; /* end of local expression */
			a->u = (const char*)-1;
			a->e = 0;

			if (s>=e)
				goto end;
			if (*s=='.') /* skip fraction in pseudo-real number */
				{++s; skipint(s,e);}
			break;
		case '!': c = _op_bno; goto unary;
		case '@': c = _op_ldw; goto unary;
		case '~': c = _op_not;
		unary:
			if (a->u==(const char*)-1)
				a->u = s; /* first in a possible row */
			if (o==_op_none)
				a->o = o = c;
			else if (o==_op_expr)
				goto end; /* came directly after alphanumeric or `)` */
			++s;
			break;
		case '+': c = _op_abs;
			if (o==_op_expr)
				{c = _op_add; goto binary;}
			goto unary;
		case '-': l_neg: c = _op_neg;
			if (o==_op_expr)
				{c = _op_sub; goto binary;}
			goto unary;
		case '%': c = _op_mdu;
			if (o==_op_mdu) /* %% */
				c = _op_mds, o = _op_expr;
			goto binary;
		case '&': c = _op_and; /* && */
			if (o==_op_and) c = _op_ban, o = _op_expr;
			goto binary;
		case '*': c = _op_mul; /* ** */
			if (o==_op_mul)
				c = _op_pow, o = _op_expr;
			goto binary;
		case '/': c = _op_dvu;
			if (o==_op_dvu) /* // */
				{l_dvs: c = _op_dvs, o = _op_expr;}
			else if (o==_op_not) /* ~/ */
				{l_sqr: c = _op_sqr; goto unary;}
			goto binary;
		case '<': c = _op_clt;
			if (o==_op_clt) /* << */
				{l_lsl: c = _op_lsl, o = _op_expr;}
			goto binary;
		case '=': c = _op_ceq;
			if (o==_op_ceq) /* == */
				c = _op_cid, o = _op_expr, ++s;
			else if (o==_op_bno) /* != */
				{l_cne: c = _op_cne, o = _op_expr;}
			else if (o==_op_clt) /* <= */
				{l_cle: c = _op_cle, o = _op_expr;}
			else if (o==_op_cgt) /* >= */
				{l_cge: c = _op_cge, o = _op_expr;}
			goto binary;
		case '>': c = _op_cgt;
			if (o==_op_clt) /* <> */
				c = _op_cni, o = _op_expr;
			else if (o==_op_lsr) /* >>> */
				c = _op_asr, o = _op_expr;
			else if (o==_op_cgt) /* >> */
				{l_lsr: c = _op_lsr, o = _op_expr;}
			goto binary;
		case '|': c = _op_ior;
			if (o==_op_ior) /* || */
				c = _op_bor, o = _op_expr;
			goto binary;
		case '^': c = _op_eor;
		binary:
			if (o!=_op_expr) /* should come after expression */
			{ /* if more than one binary in a row, discount second */
				if (o!=_op_none)
					--s;
				goto end;
			}
			a->o = o = c, ++s;
			break;
		case ',': /* explicit parameter */
		case ';': /* break */
			goto end;
		case '(':
			a->e = s;
			if (o==_op_expr)
				goto end; /* came directly after expression */
			if (++x>=sizeof(xstack)/sizeof(xstack[0])-1)
				return 0; /* stack overflow */
			/* advance and clear new frame */
			(++a)->v = 0;
			a->o = o = _op_none;
			a->u = (const char*)-1;
			a->e = 0;
			++s;
			break;
		case ')':
			if (o==_op_none || --x<0) /* empty or underflow */
				return 0;
			i = a->v;
			evaluate_chunk(--a,i,evalindir);
			a->o = o = _op_expr;
			a->u = (const char*)-1;
			a->e = 0;
			++s;
			break;
		case '_':
		CASE_ALPHABETIC
			/* look up symbol name */
			if (getvalue==0)
				goto end;
			if (a->e==0)
				a->e = s;
			/* non-integer or another alphanumeric */
			if ((p=getvalue(s,e,&i,u))==0 || o==_op_expr)
				{s = a->e; goto end;} 
			s = p, ++n;

			if (o==_op_none) /* no operator or expression; set */
				a->v = i;
			else /* evaluate */
				evaluate_chunk(a,i,evalindir);

			a->o = o = _op_expr; /* end of local expression */
			a->u = (const char*)-1;
			a->e = 0;

			if (s>=e)
				goto end;
			break;

		default:
#ifdef CHR_NE
			if ((char_type)*s==(char_type)CHR_NE) goto l_cne;
#endif
#ifdef CHR_LE
			if ((char_type)*s==(char_type)CHR_LE) goto l_cle;
#endif
#ifdef CHR_GE
			if ((char_type)*s==(char_type)CHR_GE) goto l_cge;
#endif
#ifdef CHR_RA
			if ((char_type)*s==(char_type)CHR_RA) goto l_sqr;
#endif
#ifdef CHR_SL
			if ((char_type)*s==(char_type)CHR_SL) goto l_lsl;
#endif
#ifdef CHR_SR
			if ((char_type)*s==(char_type)CHR_SR) goto l_lsr;
#endif
#ifdef CHR_MN
			if ((char_type)*s==(char_type)CHR_MN) goto l_neg;
#endif
#ifdef CHR_DV
			if ((char_type)*s==(char_type)CHR_DV) goto l_dvs;
#endif
#ifdef CHR_DS
			if ((char_type)*s==(char_type)CHR_DS) goto whitespace;
#endif
			if ((char_type)*s==(char_type)'\t') goto whitespace;
			goto end;
		}
	} while(s<e);
end:
	*v = xstack->v; /* end value should be on top of stack */
	if (--xstack->u>=str && xstack->u<(const char*)-2)
		s = xstack->u;

	return n==0 || x!=0 ? 0:s; /* got anything?  missing any parentheses? */
}

const char *parse_parameter(const char *text,
	long length,
	char *outstring,
	const char *(*getv)(const char *s, const char *e, long int *i, uint8 *u),
	uint8 *ud,
	evalvalue (*evalindir)(evalvalue))
{
	const char	*s, *e;
	evalvalue	v = 0;

	if (outstring)
		*outstring = '\0';

	if ((s=text)==0)
		return 0;
	if (length<=0)
		goto blank;
	e = length+s;
	skipwsp(s,e);

	if (s==e)
	{
	blank:
		if (outstring)
			*outstring = ' ';
		goto end;
	}

	if (*s=='\042') /* text - double quote */
		{s = quotedtext_scan(++s,e,'\042'); goto end;}
	else if (*s=='\047') /* text - single quote */
		{s = quotedtext_scan(++s,e,'\047'); goto end;}

	if ((s=evaluate_parameter(text=s,e,&v,getv,ud,evalindir))==0)
		{s = parse_wordbreak(text,e); goto end;}
	if (outstring)
		outstring[makebasestring(outstring,0,v,10,0,'-')] = 0;
end:
	return s;
}
